package ru.dbotthepony.mc.otm.item;

import net.minecraft.ChatFormatting;
import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.item.*;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.menu.FormattingHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.List;
import java.util.Objects;

@ParametersAreNonnullByDefault
@MethodsReturnNonnullByDefault
public class ItemMatterCapacitor extends Item {
	public static class ItemMatterCapacitorCapability implements ICapabilityProvider, IMatterHandler {
		private final LazyOptional<IMatterHandler> resolver = LazyOptional.of(() -> this);

		@Nonnull
		@Override
		public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
			if (cap == MatteryCapability.MATTER)
				return resolver.cast();

			return LazyOptional.empty();
		}

		private final Fraction storage;
		private final Fraction max_input;
		private final Fraction max_output;
		private final boolean is_creative;
		private Fraction energy;
		private String _energy;

		private final ItemStack stack;

		private Fraction matter() {
			CompoundTag tag = stack.getOrCreateTag();

			if (tag.contains("otm_matter")) {
				return Fraction.deserializeNBT(tag.get("otm_matter"));
			}

			return Fraction.ZERO;
		}

		private void matter(Fraction value) {
			stack.getOrCreateTag().put("otm_matter", value.serializeNBT());
		}

		public ItemMatterCapacitorCapability(ItemStack stack, Fraction storage, Fraction max_input, Fraction max_output) {
			this.stack = stack;
			is_creative = false;
			this.storage = storage;
			this.max_input = max_input;
			this.max_output = max_output;
		}

		public ItemMatterCapacitorCapability(ItemStack stack) {
			this.stack = stack;
			is_creative = true;
			storage = MatteryCapability.LONG_MAX_VALUE;
			max_input = MatteryCapability.LONG_MAX_VALUE;
			max_output = MatteryCapability.LONG_MAX_VALUE;
		}

		public ItemMatterCapacitorCapability(ItemStack stack, Fraction storage) {
			this.stack = stack;
			is_creative = false;
			this.storage = storage;
			max_input = MatteryCapability.LONG_MAX_VALUE;
			max_output = MatteryCapability.LONG_MAX_VALUE;
		}

		@Nonnull
		@Override
		public Fraction getStoredMatter() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return matter();
		}

		@Nonnull
		@Override
		public Fraction getMaxStoredMatter() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return storage;
		}

		@Nonnull
		@Override
		public Fraction getMissingMatter() {
			if (is_creative)
				return MatteryCapability.LONG_MAX_VALUE;

			return IMatterHandler.super.getMissingMatter();
		}

		@Nonnull
		@Override
		public Fraction receiveMatterOuter(Fraction howMuch, boolean simulate) {
			return receiveMatterInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction receiveMatterInner(Fraction howMuch, boolean simulate) {
			if (is_creative)
				return howMuch;

			Fraction new_matter = matter().plus(howMuch.min(max_input)).min(storage);
			Fraction diff = new_matter.minus(matter());

			if (!simulate) {
				matter(new_matter);
			}

			return diff;
		}

		@Nonnull
		@Override
		public Fraction extractMatterOuter(Fraction howMuch, boolean simulate) {
			return extractMatterInner(howMuch, simulate);
		}

		@Nonnull
		@Override
		public Fraction extractMatterInner(Fraction howMuch, boolean simulate) {
			if (is_creative)
				return howMuch;


			Fraction new_matter = matter().minus(howMuch.min(max_output)).moreThanZero();
			Fraction diff = matter().minus(new_matter);

			if (!simulate) {
				matter(new_matter);
			}

			return diff;
		}

		@Nonnull
		@Override
		public MatterDirection getDirection() {
			return MatterDirection.BIDIRECTIONAL;
		}
	}

	public final Fraction storage;

	private static final Component INFINITE_STORAGE = new TranslatableComponent("otm.item.matter.infinite").withStyle(ChatFormatting.GRAY);

	private final boolean is_creative;

	public ItemMatterCapacitor(Fraction storage) {
		super(new Properties().stacksTo(1).tab(OverdriveThatMatters.CREATIVE_TAB));
		is_creative = false;
		this.storage = storage;
	}

	public ItemMatterCapacitor() {
		super(new Properties().stacksTo(1).tab(OverdriveThatMatters.CREATIVE_TAB).rarity(Rarity.EPIC));
		is_creative = true;
		storage = MatteryCapability.LONG_MAX_VALUE;
	}

	@Override
	public void appendHoverText(ItemStack stack, @Nullable Level p_41422_, List<Component> p_41423_, TooltipFlag p_41424_) {
		super.appendHoverText(stack, p_41422_, p_41423_, p_41424_);

		if (is_creative) {
			p_41423_.add(INFINITE_STORAGE);
		} else {
			stack.getCapability(MatteryCapability.MATTER).ifPresent(cap ->
				p_41423_.add(new TranslatableComponent("otm.item.matter.normal", FormattingHelper.formatMatterValuePlain(cap.getStoredMatter()), FormattingHelper.formatMatterValuePlain(storage)).withStyle(ChatFormatting.GRAY))
			);
		}
	}

	@Nullable
	@Override
	public ICapabilityProvider initCapabilities(ItemStack stack, @Nullable CompoundTag nbt) {
		if (is_creative)
			return new ItemMatterCapacitorCapability(stack);

		return new ItemMatterCapacitorCapability(stack, storage);
	}
}
