package ru.dbotthepony.mc.otm.menu.data;

import com.mojang.blaze3d.platform.InputConstants;
import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.ClickAction;
import net.minecraft.world.inventory.ClickType;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.network.PacketDistributor;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.menu.MatteryMenu;
import ru.dbotthepony.mc.otm.network.MatteryNetworking;
import ru.dbotthepony.mc.otm.network.SetCarriedPacket;
import ru.dbotthepony.mc.otm.storage.*;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.UUID;

import static net.minecraft.client.gui.screens.Screen.hasShiftDown;

/**
 * Creates a virtual, slotless container for Player's interaction with.
 */

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class NetworkedItemView implements IStorageListener<ItemStackWrapper> {
	public final boolean remote;
	public final MatteryMenu menu;
	public final Player ply;

	protected int next_stack_id = 0;

	protected final HashMap<Integer, NetworkedItem> state = new HashMap<>();
	protected final HashMap<UUID, NetworkedItem> upstream_state = new HashMap<>();
	protected final ArrayList<Object> backlog = new ArrayList<>();

	@Nullable
	public NetworkedItem get(int id) {
		return state.get(id);
	}

	private List<NetworkedItem> view_cache;

	protected IStorageComponent<ItemStackWrapper> provider;

	public NetworkedItemView(Player ply, MatteryMenu menu, boolean remote) {
		this.remote = remote;
		this.menu = menu;
		this.ply = ply;
	}

	public void mouseClick(int index, int mouse_click_type) {
		var list = getItems();
		var action = mouse_click_type == InputConstants.MOUSE_BUTTON_LEFT ? ClickAction.PRIMARY : ClickAction.SECONDARY;
		var type = mouse_click_type == InputConstants.MOUSE_BUTTON_MIDDLE ? ClickType.CLONE : hasShiftDown() ? ClickType.QUICK_MOVE : ClickType.PICKUP;

		MatteryNetworking.send(null, new InteractPacket(menu.containerId, index >= list.size() ? -1 : list.get(index).id(), type, action));
	}

	public void setComponent(@Nullable IStorageComponent<ItemStackWrapper> provider) {
		if (provider == this.provider)
			return;

		if (this.provider != null)
			this.provider.removeListenerAuto(this);

		this.provider = provider;

		if (provider != null)
			provider.addListenerAuto(this);
	}

	public record NetworkedItem(int id, ItemStack stack, @Nullable UUID id_upstream) {
		public NetworkedItem(int id, ItemStack stack) {
			this(id, stack, null);
		}
	}

	public void clearCache() {
		view_cache = null;
	}

	public List<NetworkedItem> getItems() {
		if (view_cache != null)
			return view_cache;

		return view_cache = List.copyOf(state.values());
	}

	@Override
	public void addObject(ItemStackWrapper stack, UUID id, IStorageView<ItemStackWrapper> provider) {
		addObject(stack.stack(), id);
	}

	@Override
	public void changeObject(UUID id, Fraction new_count) {
		changeObject(id, new_count.toInt());
	}

	@Override
	public void removeObject(UUID id) {
		var get = upstream_state.get(id);

		if (get == null)
			throw new IllegalStateException("Unknown ItemStack with upstream id " + id + "!");

		upstream_state.remove(id);
		state.remove(get.id);

		if (!remote) {
			backlog.add(new StackRemovePacket(menu.containerId, get.id));
		}

		clearCache();
	}

	public void addObject(ItemStack stack, UUID id_upstream) {
		if (upstream_state.containsKey(id_upstream))
			throw new IllegalStateException("Already tracking ItemStack with upstream id " + id_upstream + "!");

		var state = new NetworkedItem(next_stack_id++, stack.copy(), id_upstream);
		this.state.put(state.id, state);
		this.upstream_state.put(id_upstream, state);

		if (!remote) {
			backlog.add(new StackAddPacket(menu.containerId, state.id, stack));
		}

		clearCache();
	}

	public void changeObject(UUID id_upstream, int new_count) {
		var get = upstream_state.get(id_upstream);

		if (get == null)
			throw new IllegalStateException("Unknown ItemStack with upstream id " + id_upstream + "!");

		get.stack.setCount(new_count);

		if (!remote) {
			backlog.add(new StackChangePacket(menu.containerId, get.id, new_count));
		}

		clearCache();
	}

	public void clear() {
		clearCache();
		upstream_state.clear();
		state.clear();

		if (!remote) {
			backlog.clear();
			backlog.add(new ClearPacket(menu.containerId));
		}
	}

	public void network() {
		if (remote)
			throw new IllegalStateException("Not a server");

		var consumer = PacketDistributor.PLAYER.with(() -> (ServerPlayer) ply);

		for (var packet : backlog) {
			MatteryNetworking.CHANNEL.send(consumer, packet);
		}

		backlog.clear();
	}

	public int calculateIOAmount(int desired_amount) {
		return desired_amount;
	}

	public void doneIOAmount(int done_amount) {

	}

	public void playerInteract(InteractPacket packet) {
		if (provider == null)
			return;

		var click = packet.click();
		var action = packet.action();
		var stack_id = packet.stack_id();

		if (click == ClickType.CLONE) {
			if (stack_id < 0 || !ply.getAbilities().instabuild)
				return;

			var get_state = get(stack_id);

			if (get_state == null) {
				return;
			}

			var copy = get_state.stack().copy();
			copy.setCount(Math.min(copy.getCount(), copy.getMaxStackSize()));

			ply.containerMenu.setCarried(copy);
			MatteryNetworking.send((ServerPlayer) ply, new SetCarriedPacket(ply.containerMenu.getCarried()));
			ply.containerMenu.setRemoteCarried(ply.containerMenu.getCarried().copy());
			return;
		}

		if (click == ClickType.QUICK_MOVE && stack_id > -1) {
			var get_state = get(stack_id);

			if (get_state == null) {
				return;
			}

			int amount = calculateIOAmount(action == ClickAction.PRIMARY ? get_state.stack().getMaxStackSize() : Math.max(1, get_state.stack().getMaxStackSize() / 2));
			var extracted = provider.extractStack(get_state.id_upstream(), new Fraction(amount), true);

			if (!extracted.isEmpty()) {
				var move = menu.quickMoveToInventory(extracted.stack(), false);

				if (move.remaining().getCount() != extracted.stack().getCount()) {
					provider.extractStack(get_state.id_upstream(), new Fraction(extracted.stack().getCount() - move.remaining().getCount()), false);
					doneIOAmount(extracted.stack().getCount() - move.remaining().getCount());
				}
			}

			return;
		}

		if (!menu.getCarried().isEmpty() && click != ClickType.QUICK_MOVE) {
			// try to put
			if (action == ClickAction.PRIMARY) {
				var carried = menu.getCarried();
				int amount = calculateIOAmount(carried.getCount());

				if (amount == carried.getCount()) {
					var leftover = provider.insertStack(new ItemStackWrapper(menu.getCarried()), false);
					menu.setCarried(leftover.stack());
					doneIOAmount(amount - leftover.stack().getCount());

					MatteryNetworking.send((ServerPlayer) ply, new SetCarriedPacket(menu.getCarried()));
					menu.setRemoteCarried(menu.getCarried().copy());
				} else if (amount != 0) {
					var copy = carried.copy();
					copy.setCount(amount);

					var leftover = provider.insertStack(new ItemStackWrapper(copy), false);
					doneIOAmount(amount - leftover.stack().getCount());

					leftover.setCount(carried.getCount() - amount + leftover.getCountInt());
					menu.setCarried(leftover.stack());

					MatteryNetworking.send((ServerPlayer) ply, new SetCarriedPacket(menu.getCarried()));
					menu.setRemoteCarried(menu.getCarried().copy());
				}
			} else {
				var copy = menu.getCarried().copy();
				copy.setCount(1);

				if (calculateIOAmount(1) == 1 && provider.insertStack(new ItemStackWrapper(copy), false).isEmpty()) {
					menu.getCarried().shrink(1);
					doneIOAmount(1);
					MatteryNetworking.send((ServerPlayer) ply, new SetCarriedPacket(menu.getCarried()));
					menu.setRemoteCarried(menu.getCarried().copy());
				}
			}
		} else if (stack_id > -1) {
			var get_state = get(stack_id);

			if (get_state == null) {
				return;
			}

			int amount = calculateIOAmount(action == ClickAction.PRIMARY ? get_state.stack().getMaxStackSize() : Math.max(1, get_state.stack().getMaxStackSize() / 2));

			var extracted = provider.extractStack(get_state.id_upstream(), new Fraction(amount), false);

			doneIOAmount(extracted.getCountInt());
			menu.setCarried(extracted.stack());
			MatteryNetworking.send((ServerPlayer) ply, new SetCarriedPacket(menu.getCarried()));
			menu.setRemoteCarried(menu.getCarried().copy());
		}
	}
}
